
def standardSetup():
    import GangaTest.PACKAGE
    PACKAGE.standardSetup()


def loadPlugins( config = None ):
    import GangaTest.Framework.runner
    import GangaTest.Lib.GListApp
    import GangaTest.Lib.TFile
    import GangaTest.Lib.TestApplication
    import GangaTest.Lib.TestSubmitter # TestSubmitter, TestSplitter here
    import GangaTest.Lib.TestRobot
    import GangaTest.Lib.TestObjects
    import GangaTest.Framework.runner


