# Import classes that should be in the Tasks namespace
## Import the list of tasks and the task and abstract job definition

## Tasks
from NA62Task import NA62Task
from NA62Transform import NA62Transform
from NA62Unit import NA62Unit

