from __future__ import absolute_import

from GangaCore.Core.exceptions import PluginError
raise PluginError("The GangaLSST module has not been upgraded for python 3. The last python 2 Ganga version is 7.1.15 . Please contact the ganga devs to discuss updating this module.")

from . import Lib

