# File: GangaTutorial/__init__.py

def loadPlugins( config = {} ):

    from .Lib import PrimeFactorizer
    from .Lib import PrimeFactorizerSplitter
    from .Lib import PrimeTableDataset
	
    return None
